/*!
 * Contro
 * (c) 2017 Niklas Higi
 * Released under the MIT License.
 */
(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (factory((global.Contro = {})));
}(this, (function (exports) { 'use strict';

  class Vector2 {
      constructor(x = 0, y = 0) {
          this.x = x;
          this.y = y;
      }
  }

  const mouseButtons = ['left', 'middle', 'right'];
  class Mouse {
      constructor({ canvas, doc = document }) {
          this.pointerLocked = false;
          this.pointerMovement = new Vector2();
          this.pressedButtons = new Set();
          this.queuedButtons = new Set();
          this.scrollDistance = 0;
          this.canvas = canvas;
          this.document = doc;
          const on = this.canvas.addEventListener.bind(this.canvas);
          on('mousedown', (event) => {
              store.preferGamepad = false;
              this.pressedButtons.add(event.button);
              this.queuedButtons.add(event.button);
          });
          on('mouseup', (event) => {
              store.preferGamepad = false;
              this.pressedButtons.delete(event.button);
              this.queuedButtons.delete(event.button);
          });
          on('mousemove', (event) => {
              store.preferGamepad = false;
              this.pointerMovement.x += event.movementX;
              this.pointerMovement.y += event.movementY;
          });
          on('wheel', (event) => {
              store.preferGamepad = false;
              const distance = event.deltaY;
              this.scrollDistance += distance;
          });
      }
      parseButton(button) {
          if (typeof button === 'string') {
              if (mouseButtons.includes(button)) {
                  return mouseButtons.indexOf(button);
              }
              else {
                  throw new Error(`There is no mouse button called "${button}"!`);
              }
          }
          else {
              if (button < mouseButtons.length) {
                  return button;
              }
              else {
                  throw new Error(`There is no mouse button with the index ${button}!`);
              }
          }
      }
      button(button) {
          const that = this;
          button = this.parseButton(button);
          return {
              label: ['Left', 'Middle', 'Right'][button] + ' Mouse Button',
              query() {
                  button = that.parseButton(button);
                  if (!this.hasOwnProperty('trigger')) {
                      if (that.queuedButtons.has(button)) {
                          that.queuedButtons.delete(button);
                          return true;
                      }
                      return false;
                  }
                  else {
                      return that.pressedButtons.has(button);
                  }
              },
              get trigger() {
                  delete this.trigger;
                  return this;
              },
          };
      }
      pointer() {
          return {
              label: 'Cursor',
              query: () => {
                  const movement = this.pointerMovement;
                  this.pointerMovement = new Vector2(0, 0);
                  return movement;
              },
          };
      }
      wheel() {
          return {
              label: 'Mouse wheel',
              query: () => {
                  const distance = this.scrollDistance;
                  this.scrollDistance = 0;
                  return distance;
              },
          };
      }
      lockPointer() {
          this.canvas.requestPointerLock();
      }
      unlockPointer() {
          this.document.exitPointerLock();
      }
      isPointerLocked() {
          return this.document.pointerLockElement === this.canvas;
      }
  }

  /**
   * A map of all the supported key values (property names) and their respective
   * aliases (property values)  that can be used with the `Keyboard` class. The
   * first alias for each key value will be used as a label.
   */
  const keyMap = {
      ' ': ['Space', 'Spacebar', 'Space Bar'],
      'AltGraph': ['Alt Gr'],
      'ArrowDown': ['Down'],
      'ArrowLeft': ['Left'],
      'ArrowRight': ['Right'],
      'ArrowUp': ['Up'],
      'Backspace': ['Backspace'],
      'Control': ['Ctrl', 'Ctl'],
      'Delete': ['Delete', 'Del'],
      'Enter': ['Enter', 'Return'],
      'Escape': ['Escape', 'Esc'],
      'Insert': ['Insert', 'Ins'],
      'PageDown': ['Page Down', 'PgDown'],
      'PageUp': ['Page Up', 'PgUp'],
      'Tab': ['Tab'],
  };
  function findKeyValue(keyString) {
      if (keyString.length === 1)
          return keyString.toLowerCase();
      Object.keys(keyMap).forEach(keyValue => {
          keyMap[keyValue].forEach(key => {
              if (keyString.toLowerCase() === key.toLowerCase()) {
                  keyString = keyValue;
              }
          });
      });
      return keyString;
  }
  function getKeyLabel(key) {
      return key in keyMap ? keyMap[key][0] : (key.length === 1 ? key.toUpperCase() : key);
  }

  const arrowKeyTemplates = {
      arrows: ['Arrow keys', ['ArrowUp', 'ArrowLeft', 'ArrowDown', 'ArrowRight']],
      wasd: ['WASD', ['W', 'A', 'S', 'D']],
  };
  class Keyboard {
      constructor(
          { doc = document } = {}) {
          this.pressedKeys = new Set();
          this.queuedKeys = new Set();
          this.document = doc;
          this.document.addEventListener('keydown', (event) => {
              store.preferGamepad = false;
              let key = event.key;
              if (key === key.toUpperCase())
                  key = key.toLowerCase();
              this.pressedKeys.add(key);
              this.queuedKeys.add(key);
              return false;
          });
          this.document.addEventListener('keyup', (event) => {
              store.preferGamepad = false;
              let key = event.key;
              if (key === key.toUpperCase())
                  key = key.toLowerCase();
              this.pressedKeys.delete(key);
              this.queuedKeys.delete(key);
              return false;
          });
      }
      key(key) {
          const that = this;
          key = findKeyValue(key);
          return {
              label: getKeyLabel(key),
              query() {
                  return this.hasOwnProperty('trigger') ? that.pressedKeys.has(key) : that.queuedKeys.delete(key);
              },
              get trigger() {
                  delete this.trigger;
                  return this;
              },
          };
      }
      directionalKeys(keys, label) {
          let defaultLabel;
          if (typeof keys === 'string') {
              keys = keys.toLowerCase();
              if (keys in arrowKeyTemplates) {
                  const template = arrowKeyTemplates[keys.toLowerCase()];
                  defaultLabel = template[0];
                  keys = template[1];
              }
              else {
                  throw new Error(`Directional key template "${keys}" not found!`);
              }
          }
          else {
              if (keys.length === 4) {
                  keys = keys.map(key => findKeyValue(key));
                  defaultLabel = keys.map(key => getKeyLabel(key)).join('');
              }
              else {
                  throw new Error('Directional key templates have to consist of four keys!');
              }
          }
          return {
              label: label || defaultLabel,
              query: () => {
                  const vector = new Vector2();
                  if (this.key(keys[0]).query())
                      vector.y -= 1;
                  if (this.key(keys[1]).query())
                      vector.x -= 1;
                  if (this.key(keys[2]).query())
                      vector.y += 1;
                  if (this.key(keys[3]).query())
                      vector.x += 1;
                  return vector;
              },
          };
      }
  }

  /**
   * A map of all the supported button numbers (array indices) and their
   * respective aliases (the strings in the arrays) that can be used with the
   * `Gamepad` class. The first alias for each button will be used as a label.
   */
  const buttonMap = [
      ['A'],
      ['B'],
      ['X'],
      ['Y'],
      ['Left Bumper', 'LB'],
      ['Right Bumper', 'RB'],
      ['Left Trigger', 'LT'],
      ['Right Trigger', 'RT'],
      ['Back', 'View'],
      ['Start'],
      ['Left Stick'],
      ['Right Stick'],
      ['Up', 'DpadUp'],
      ['Down', 'DpadDown'],
      ['Left', 'DpadLeft'],
      ['Right', 'DpadRight'],
      ['Home', 'Guide', 'Xbox'],
  ];
  function findButtonNumber(button) {
      if (typeof button === 'number')
          return button;
      let buttonNumber = 0;
      for (const buttonAliases of buttonMap) {
          for (const buttonAlias of buttonAliases) {
              if (button.toLowerCase() === buttonAlias.toLowerCase()) {
                  return buttonNumber;
              }
          }
          buttonNumber++;
      }
  }
  function getButtonLabel(button) {
      return buttonMap[button][0];
  }

  const gamepadSticks = {
      left: { xAxis: 0, yAxis: 1 },
      right: { xAxis: 2, yAxis: 3 },
  };
  class Gamepad {
      constructor(
          { win = window, nav = navigator } = {}) {
          this.pressedButtons = new Set();
          this.gamepadTimestamp = 0;
          this.window = win;
          this.navigator = nav;
          this.window.addEventListener('gamepadconnected', ({ gamepad }) => {
              if (!this.isConnected()) {
                  if (gamepad.mapping === 'standard') {
                      this.gamepadIndex = gamepad.index;
                      store.preferGamepad = true;
                  }
              }
          });
          this.window.addEventListener('gamepaddisconnected', ({ gamepad }) => {
              if (this.gamepadIndex === gamepad.index) {
                  this.gamepadIndex = undefined;
                  store.preferGamepad = false;
              }
          });
      }
      isConnected() {
          return this.gamepadIndex !== undefined && this.gamepad.connected;
      }
      get gamepad() {
          const gamepad = this.navigator.getGamepads()[this.gamepadIndex];
          if (gamepad.timestamp > this.gamepadTimestamp)
              store.preferGamepad = true;
          this.gamepadTimestamp = gamepad.timestamp;
          return gamepad;
      }
      button(button) {
          const that = this;
          const buttonNumber = findButtonNumber(button);
          return {
              label: getButtonLabel(buttonNumber),
              fromGamepad: true,
              query() {
                  if (!that.isConnected())
                      return false;
                  if (!this.hasOwnProperty('trigger')) {
                      if (that.gamepad.buttons[buttonNumber].pressed) {
                          if (!that.pressedButtons.has(buttonNumber)) {
                              that.pressedButtons.add(buttonNumber);
                              return true;
                          }
                      }
                      else {
                          that.pressedButtons.delete(buttonNumber);
                      }
                      return false;
                  }
                  else {
                      return that.gamepad.buttons[buttonNumber].pressed;
                  }
              },
              get trigger() {
                  delete this.trigger;
                  return this;
              },
          };
      }
      stick(stick) {
          if (typeof stick === 'string') {
              if (stick in gamepadSticks) {
                  stick = gamepadSticks[stick];
              }
              else {
                  throw new Error(`Gamepad stick "${stick}" not found!`);
              }
          }
          return new Vector2(this.gamepad.axes[stick.xAxis], this.gamepad.axes[stick.yAxis]);
      }
  }

  function and(...controls) {
      if (controls.length < 2)
          throw new Error('Less than two controls specified!');
      return {
          label: controls.map(control => control.label).join(' + '),
          query: () => {
              for (const control of controls) {
                  if (!control.query())
                      return false;
              }
              return true;
          },
      };
  }

  function or(...controls) {
      if (controls.length < 2)
          throw new Error('Less than two controls specified!');
      return {
          get label() {
              return (controls.filter(control => control.fromGamepad).length === 0 ?
                  controls[0]
                  : store.preferGamepad ?
                      controls.filter(control => control.fromGamepad === true)[0]
                      : controls.filter(control => control.fromGamepad !== true)[0]).label;
          },
          query: () => {
              let sampleQueryValue;
              for (const control of controls) {
                  const queryValue = control.query();
                  sampleQueryValue = queryValue;
                  if (queryValue)
                      return queryValue;
              }
              if (typeof sampleQueryValue === 'boolean')
                  return false;
          },
      };
  }

  let store = {
      preferGamepad: false,
  };

  exports.store = store;
  exports.Mouse = Mouse;
  exports.Keyboard = Keyboard;
  exports.Gamepad = Gamepad;
  exports.and = and;
  exports.or = or;

  Object.defineProperty(exports, '__esModule', { value: true });

})));
